from microprediction.writer import MicroWriter
from microprediction.conventions import api_url
import time
from apscheduler.schedulers.blocking import BlockingScheduler
from pprint import pprint
import datetime

class MicroPoll(MicroWriter):

    #--------------------------------------------
    #  You may wish to override these methods
    #--------------------------------------------

    def logger(self, data):
        """ Called after each attempt to poll data and send it """
        self.recent.append(data)
        if len(self.recent)>100:
            self.recent = self.recent[-100:]
        if self.verbose:
            pprint(data)
            print(' ',flush=True)

    def downtime(self):
        """ Also called after each attempt to send data """
        # By default it does a little MUID mining, or a little more if the balance is sinking
        self.maybe_bolster_balance_by_mining()


    def determine_next_value(self, source_value):
        """ Receives source data and decides what to send, if anything """
        # This should return a float or None
        return float(source_value) if source_value is not None else None

    #--------------------------------------------
    #  Probably don't want to override the rest
    #--------------------------------------------


    def __init__(self, name, func, interval, write_key="invalid_key", base_url=None, verbose=True, func_args=None, **kwargs):
        """  Create a stream by polling every 20 minutes, say
            param: name  str   stream name ending in .json
            func:        function    returns float  (data feed function)
            interval:    int     minutes between polls
            func_args    dict    optional dict of arguments to be passed to func
        """
        assert self.is_valid_name(name),'name not valid'
        super().__init__(base_url=base_url or api_url(),write_key=write_key,verbose=verbose)
        self.name = name
        self.interval = interval
        self.func = func
        self.func_args = func_args
        self.recent = list()
        self.mining_time = 0
        self.mining_success = 0
        self.test_value = self.call_func()  # Fail fast
        print('Created poller. Example value '+str(self.test_value))

    def __repr__(self):
        self_data= {'name':self.name,'func':str(self.func),'interval':self.interval,'mining_time':self.mining_time}
        if self.func_args is not None:
            self_data.update(self.func_args)
        return self_data

    def call_func(self):
        if isinstance(self.func_args,list):
            return self.func(*self.func_args)
        elif isinstance(self.func_args, dict):
            return self.func(**self.func_args)
        else:
            return self.func()

    def task(self):
        """ Scheduled task that runs every minute """
        start_time = time.time()
        data = {'start_time': str(datetime.datetime.now())}
        try:
            source_value = self.call_func()
        except Exception as e:
            source_value = None
        next_value   = self.determine_next_value(source_value)
        data.update({'source_value':source_value,
                     'next_value':next_value,
                     'elapsed after polling':time.time()-start_time})
        res = self.touch(self.name) if next_value is None else self.set(name=self.name, value=next_value)
        data.update({'value': next_value, "res": res,'elapsed after sending':time.time()-start_time})
        self.logger(data=data )
        self.downtime()
        data.update({'elapsed after downtime':time.time()-start_time})

    def maybe_bolster_balance_by_mining(self):
        """ Mine just a littel to avoid stream dying due to bankruptcy """
        balance = self.get_balance()
        if balance < 0:
            muid_time = time.time()
            key = self.bolster_balance_by_mining(seconds=max(50, int(abs(balance) / 10)))
            mining_time = time.time() - muid_time
            self.mining_time += mining_time
            if key:
                print('************************')
                print('     FOUND MUID !!!     ')
                print('************************', flush=True)
                self.mining_success += 1

    def run(self):
        data = {'type': 'scheduler', 'scheduler start time': time.time()}
        data.update(self.__repr__())
        self.logger(data=data)
        scheduler = BlockingScheduler()
        scheduler.add_job(self.task, 'interval', minutes=self.interval, max_instances=30)
        try:
            scheduler.start()
        except (KeyboardInterrupt, SystemExit):
            pass
        data.update({'stopping time':time.time()})
        self.logger(data=data)


class ChangePoll(MicroPoll):

    #    Illustrates predicting the change in a quantity ... when feed is not entirely reliable
    #    If we don't get good data from the feed or it appears to be stale, we don't publish the change

    COLD = 0
    WARM = 1

    def __init__(self, name, func, interval, write_key, func_args=None):
        super().__init__(name=name, func=func, interval=interval, func_args=func_args, write_key=write_key )
        self.prev_value = None
        self.feed_state = ChangePoll.COLD
        self.current_value = None

    def alert(self,message):
        print(message, flush=True)

    def determxine_next_value(self, source_value):
        """ Returns a change if feed is warm, else None """

        if self.feed_state == ChangePoll.WARM:
            # A warm state means that previous speed exists and is not stale
            assert self.prev_value is not None
            current_value = source_value
            if current_value is None:
                self.feed_state = "cold"
                self.alert(message='Something amiss with feed')
                return None
            else:
                current_value = float(source_value)
                value_change = float(current_value) - float(self.prev_value)
                if abs(value_change) < 1e-5:
                    self.feed_state = "cold"  # Feed is stale, don't judge
                    self.logger({'type':'feed_status','message':"****  Feed unchanged at " + str(datetime.datetime.now())})
                else:
                    self.prev_value = current_value
                    return value_change
        elif self.feed_state == ChangePoll.COLD:
            # Wait until feed is back up and speeds start changing
            prev_prev = self.prev_value if self.prev_value else None
            prev_speed = source_value
            if (prev_speed is not None) and (prev_prev is not None) and abs(
                    float(prev_prev) - float(prev_speed)) > 1e-5:
                self.feed_state = ChangePoll.WARM
                self.logger({'type':'feed_status','message':'**** Feed resumed at ' + str(datetime.datetime.now())})
            return None





import logging
import requests
import urllib
import time
import pytz
import random
from datetime import datetime
from pprint import  pprint

from microprediction import MicroWriter
from microprediction.config_private import TRAFFIC_WRITE_KEY_13

from apscheduler.schedulers.blocking import BlockingScheduler


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def wait_between_attempts():
    """ Incremental backoff between connection attempts """
    wait_time = 19.3  # time is in seconds
    while True:
        yield wait_time
        wait_time = min(wait_time * 1.5, 30)
        wait_time *= (100 + random.randint(0, 50)) / 100

wait_time = wait_between_attempts()


def fetch_live_data():
    while True:
        try:
            file = urllib.request.urlopen("https://www.ndbc.noaa.gov/data/realtime2/WPOW1.cwind")
            if file.status==200:
                # NOTE: file is not indexable
                for i, line in enumerate(file):
                    # most recent data is at line number 3 in the file
                    if i == 2:
                        data = line.decode("utf-8").split()
                        direction = float(data[5]) / 360        # normalize between [0, 1)
                        speed = float(data[6]) / 10             # attempt to normalize
                        return speed, direction
        except:
            logger.error("Connection error: reconnecting...")
            time.sleep(next(wait_time))
    return speed, direction


NAME_SPEED = 'seattle_wind_speed.json'
NAME_DIR   = 'seattle_wind_direction.json'

initial_speed, initial_dir = fetch_live_data()
print("Initial speed is " + str(initial_speed) + " m/s and initial direction is " + str(initial_dir), flush=True)

try:
    mw = MicroWriter(write_key=TRAFFIC_WRITE_KEY_13)
except:
    raise Exception("You need to set the write key for this example to work")


def poll_and_send():
    utc_now = pytz.utc.localize(datetime.utcnow())
    pst_now = utc_now.astimezone(pytz.timezone("America/Los_Angeles"))
    speed, direction = fetch_live_data()
    res = mw.cset(names=[NAME_SPEED, NAME_DIR],values=[speed, direction])
    pprint({'PST time':pst_now.strftime("%H:%M"),'speed':speed,'direction':direction,"res":res})
    print('',flush=True)

def run():
    print('Starting scheduler',flush=True)
    scheduler = BlockingScheduler()
    scheduler.add_job(poll_and_send, 'interval', minutes=1)
    try:
        scheduler.start()
    except (KeyboardInterrupt, SystemExit):
        pass
    print('Stopping scheduler',flush=True)


if __name__=="__main__":
    run()




# Only used for testing

from microprediction import MicroCrawler
import time, datetime
import numpy as np
from pprint import pprint

class DevTestingCrawler(MicroCrawler):

        " Crawler used for testing code releases "

        def __init__(self, write_key, pass_callback, fail_callback):
            super().__init__(stop_loss=2, min_lags=0, sleep_time=1, write_key=write_key, quietude=10, verbose=False)
            self.pass_callback = pass_callback
            self.fail_callback = fail_callback
            self.initial_balance = self.get_balance()

        def setup(self,**kwargs):
            """ Return error messages if any """
            pass

        def teardown(self,**kwargs):
            """ Return error messages if any """
            pass

        def candidate_streams(self,**ignore):
            """ Quickly quickly """
            return [name for name, sponsor in self.get_sponsors().items() if not '~' in name]

        def candidate_delays(self,**ignore):
            """ Fail fast """
            return [70]

        def sample(self, lagged_values, lagged_times=None, **ignore):
            if len(lagged_values or []) > 5:
                return super().sample(lagged_values=lagged_values, lagged_times=lagged_times)
            else:
                return sorted(
                    np.random.randn(self.num_predictions))


        def run_dev_tests(self,timeout=180,name='devtest_crawler'):
            """ Returns error report in form of dict """
            report = {'crawler': name, 'timeout': timeout, 'start_time': time.time(),
                      'start_datetime': str(datetime.datetime.now())}
            print(self.write_key)

            # Initial checks
            try:
                setup_errors = self.setup()
            except Exception as e:
                setup_errors = {'error':'setup error','message':str(e)}

            if setup_errors is None:
                # Run checks
                try:
                    self.run(timeout=timeout)
                    run_errors = None
                except Exception as e:
                    print(str(e), flush=True)
                    passed = False
                    run_errors = {'error':'run error','message':str(e)}
            else:
                run_errors = None

            if run_errors is None:
                # Teardown checks
                try:
                    teardown_errors = self.setup()
                except Exception as e:
                    teardown_errors = {'error': 'error thrown by teardown', 'message': str(e)}
            else:
                teardown_errors = None

            any_errors = setup_errors or run_errors or teardown_errors
            if setup_errors:
                report.update(setup_errors)
            if run_errors:
                report.update(run_errors)
            if teardown_errors:
                report.update(teardown_errors)

            if any_errors and self.pass_callback is not None:
                successfully_reported_pass = self.pass_callback(report)
                if successfully_reported_pass==False:
                    report.update({'reporting_failure':True})
                    self.fail_callback(report)
            if any_errors and (self.fail_callback is not None):
                self.fail_callback(report)
            pprint(report)


if __name__=="__main__":
    try:
        from microprediction.config_private import pass_callback, fail_callback, FLAMMABLE_COD
    except ImportError as e:
        pass_callback=None
        fail_callback=None

    crawler = DevTestingCrawler(write_key=FLAMMABLE_COD,pass_callback=pass_callback,fail_callback=fail_callback)
    crawler.run_dev_tests(timeout=100,name='local test')



# Provides an example of a self-navigating algorithm which you are under no obligation to use
from microprediction.writer import MicroWriter
from microconventions import api_url
import random, time
from collections import OrderedDict
from microprediction.samplers import exponential_bootstrap, approx_mode
import pprint
import numpy as np
from statistics import median


class MicroCrawler(MicroWriter):
    """To create a crawler you can derive from this class and it is suggested you overwrite one of more of the following methods:

           sample            - Method that creates predictions
           downtown          - Stuff to do when there is nothing terribly urgent

        You might also want to change

           candidate_streams  - A list of streams you want your crawler to look at
           candidate_delays   - A list of horizons to predict (from the set self.DELAYS)
           update_frequency   - How many times to predict per arriving data point

       Good luck!
    """


    #############################################################
    #   Override these methods                                  #
    #############################################################


    def candidate_streams(self):
        """ Modify this as you see fit to select streams for your crawler
            For example your crawler might like z streams or not, or might require
            a minimum number of lags in the time series.

             :return [ str ]    List of stream names that your bot would be willing to participate in

        """
        budgets  = self.get_budgets()
        sponsors = self.get_sponsors()
        # Stream criteria (combines with AND)
        not_too_dull = [name for name, budget in budgets.items() if float(budget) >= self.min_budget]
        not_too_competitive = [name for name, budget in budgets.items() if float(budget) <= self.max_budget]
        well_sponsored = [name for name, sponsor in sponsors.items() if
                          len(sponsor.replace(' ', '')) >= float(self.sponsor_min)]
        inclusion_criteria = [not_too_dull, not_too_competitive, well_sponsored]
        names = list(set.intersection(*map(set, inclusion_criteria)))
        return names

    def candidate_delays(self, name=None):
        """  Determines how far ahead your bot is willing to predict.  However only valid delay times are acceptable (e.g 70)

            :returns [ int ]    List of horizons your bot is willing to participate in
        """
        return self.DELAYS

    def update_frequency(self, name=None, delay=None):
        """ Determines how often to update samples """
        return 20 if "~" in name else 1 

    def sample(self, lagged_values, lagged_times=None, name=None, delay=None):
        """ An example of a sample method. This is where all the intelligence goes

               :param lagged_times [ float ]    Vector with most recent listed first
               :param lagged_times [ float ]    Vector of epoch times, most recent listed first
               :param name          str         Name of stream
               :param delay         int         Prediction horizon in seconds

         Swap this code out for whatever you choose to do.
         Return a vector of numbers somewhat indicative of a probability density, of length self.num_predictions

        """
        scenarios = exponential_bootstrap(lagged=lagged_values, num=self.num_predictions, decay=0.01)

        assert len(scenarios)==self.num_predictions, "Your sammpler should product a vector of length "+str(self.num_predictions)
        return sorted(scenarios)

    def downtime(self,seconds,**ignored):
        """ This will be called when a small gap opens up between times when you need to predict """
        # Be careful if you override it or you might be late for your next prediction
        time.sleep(seconds)

    def withdrawing(self,horizon):
        """ Teardown when you withdraw from a horizon """
        pass

    #############################################################
    #   Typically don't want to override the rest of these      #
    #############################################################


    def __init__(self, write_key=None, base_url=None, verbose=False, quietude=50, stop_loss=10, min_budget=0., max_budget=10, min_lags = 25, max_lags=1000000, sponsor_min=12, sleep_time=300):
        """  """
        super().__init__(base_url=base_url or api_url(), write_key=write_key, verbose=verbose )
        assert self.key_difficulty(write_key) >= 7, "Invalid write_key for crawler. See www.muid.org to mine one. "
        assert not self.base_url[-1]=='/','Base url should not have trailing /'
        assert stop_loss>0,"Stop loss must be positive "
        self.quietude    = int(quietude)       # e.g. if set to 10, will only print 1/10th of the time
        self.sponsor_min = sponsor_min         # Only choose streams with sponsors at least this long
        self.sleep_time  = sleep_time          # Minimum number of seconds to add between predictions on a given stream
        self.stop_loss   = stop_loss           # How much to lose before giving up on a stream
        self.min_budget  = min_budget          # Play with highly competitive algorithms?
        self.max_budget  = max_budget          # Play with highly competitive algorithms?
        self.min_lags    = min_lags            # Insist on historical data
        self.max_lags    = max_lags
        self.performance = self.get_performance()
        self.active      = self.get_active()   # List of active horizons
        self.horizon_blacklist = list()
        self.cancelled   = list()
        self.next_prediction_time = dict()
        self.backoff     = dict()              # Lookup of times to sleep when feeds are likely stale
        self.message_log = list()
        self.seconds_until_next = 10           # Set just before self.sample() is called so we know how long we have
        self.last_performance_check = time.time()
        self.last_new_horizon = time.time()
        self.stream_candidates = self.candidate_streams()
        self.next_prediction_time = dict()     # Indexed by horizon

    def feel_like_talking(self):
        return self.quietude<=1 or random.choice(range(self.quietude))==0

    def upcoming(self,num=10, relative=True):
        """ List horizons and seconds until next prediction is required  """
        soon = sorted(self.next_prediction_time.items(), key=lambda d: d[1])[:num]
        if relative:
            t = time.time()
            return [ (n,s-t) for n,s in soon]
        else:
            return soon

    def update_seconds_until_next(self,exclude=None):
        exclude = exclude or list()
        upcoming = [t for h, t in self.upcoming(num=2, relative=True) if not h in exclude]
        try:
            self.seconds_until_next = min(upcoming)
        except:
            self.seconds_until_next = 10*60
        seconds_until_quit = self.end_time-time.time()
        if int(seconds_until_quit) in [1000,100,10,5]:
            print('Seconds until quit: '+str(seconds_until_quit),flush=True)
        self.seconds_until_next = min(self.seconds_until_next,seconds_until_quit)
        return self.seconds_until_next

    def __repr__(self):
        return {'quietude':self.quietude,
                'sponsor_min':self.sponsor_min,
                'stop_loss':self.stop_loss,
                'min_budget':self.min_budget,
                'max_budget':self.max_budget,
                'min_lags':self.min_lags,
                'max_lags':self.max_lags,
                'num_active':len(self.active),
                'num_blacklisted':len(self.horizon_blacklist),
                'current_balance':self.get_balance(),
                'recently_erroneous':self.get_errors()[-3:],
                'currently_worst': self.worst_active_horizons()[:3],
                'recently_cancelled':self.cancelled[:3],
                'recently_blacklisted': self.horizon_blacklist[-3:],
                'backoff':sorted(self.backoff.items(),key=lambda d: d[1])[:3],
                'upcoming':self.upcoming(num=3,relative=True)}

    def recent_updates(self):
        r = self.__repr__()
        return dict( [ (k,v) for k,v in r.items() if ('recent' in k or 'current' in k) ] )

    def active_performance(self, reverse=False ):
        return OrderedDict( sorted( [ (horizon, balance) for horizon, balance in self.performance.items() if horizon in self.active], key = lambda e: e[1], reverse=reverse ) )

    def worst_active_horizons(self):
        return [ horizon for horizon, balance in self.active_performance(reverse=True).items() if balance<-abs(self.stop_loss) ]

    def withdraw(self,horizon):
        name, delay = self.split_horizon_name(horizon)
        self.cancel(name=name, delays=[delay])
        horizon = self.horizon_name(name=name, delay=delay)
        if horizon in self.next_prediction_time:
            del self.next_prediction_time[horizon]
        self.cancelled.append(horizon)
        print("Withdrawing from " + horizon, flush=True)
        self.horizon_blacklist.append(horizon)
        self.active      = self.get_active()
        self.performance = self.get_performance()
        self.withdrawing(horizon)

    def withdraw_from_worst(self, num=1):
        horizons = self.worst_active_horizons()[:num]
        for horizon in horizons:
            self.withdraw(horizon=horizon)

    def next_horizon(self, exclude=None):
        """ Choose an urgent horizon """
        candidates = self.stream_candidates
        if exclude:
            candidates = [ c for c in candidates if not c in exclude ]
        if candidates:
            horizons = list()
            for name in candidates:
                delays = self.candidate_delays(name=name)
                for delay in delays:
                    horizon = self.horizon_name(name=name,delay=delay)
                    if not horizon in self.active:
                        if not horizon in self.horizon_blacklist:
                            if not (horizon in self.next_prediction_time) or time.time()>self.next_prediction_time[horizon]:
                                horizons.append(horizon)
            if horizons:
                return random.choice(horizons)

    def estimate_prediction_time(self, lagged_times, delay, num_intervals):
        # Base on lagged times of observations and the horizon, set a time at which to next create a prediction
        if len(lagged_times) > 5:
            neg_dt = approx_mode(np.diff(lagged_times))
            dt = -neg_dt if neg_dt is not None else None
            if dt is None or self.sleep_time > dt:
                dt = self.sleep_time
            # e.g. dt=90 and delay=310  we have 80 seconds grace
            expected_at = lagged_times[0] + dt * num_intervals
        else:
            expected_at = time.time() + 600
            dt = self.sleep_time

        if delay >= dt:
            earliest = expected_at + 5
            latest = expected_at + delay % dt - 5
        else:
            latest = expected_at - dt - 5
            earliest = expected_at - 2 * dt + 5
        if earliest > latest:
            earliest = latest
        predict_time = np.random.rand() * (latest - earliest) + earliest
        return predict_time, dt, earliest, latest, expected_at

    def predict_and_submit(self, name, delay, lagged_times ):
        """ Given a stream and horizon, try to submit predictions """
        horizon = self.horizon_name(name=name, delay=delay)
        lagged_values = self.get_lagged_values(name)
        execut = 0
        message = {'name':name,'delay':delay,'submitted':False}
        if len(lagged_values) > self.max_lags:
            message.update({"reason": "Too many lags - blacklisting", "lagged_len": len(lagged_values)})
            self.horizon_blacklist.append(horizon)
        if len(lagged_values or []) < self.min_lags:
            message.update({"reason": "Too few lags", "lagged_len": len(lagged_values)})
            if len(lagged_times)>5:
                dt = (lagged_times[1]-lagged_times[-1] )/(len(lagged_times)-1)
                next_predict_time = self.min_lags*dt + lagged_times[-1]
                if next_predict_time<time.time():
                    next_predict_time = time.time()+10*60*60
                    if not horizon in self.backoff:
                        self.backoff[horizon] = 10*60*60
            else:
                next_predict_time = time.time()+5*60
            self.next_prediction_time[horizon]=next_predict_time
            message.update({"next":next_predict_time,"in":next_predict_time-time.time()})
        else:
            # Call self.sample() but first let it know how much time it has
            self.update_seconds_until_next(exclude=[horizon])
            scenario_values = self.sample(lagged_values=lagged_values,lagged_times=lagged_times, name=name, delay=delay )
            execut = self.submit(name=name, values=scenario_values, delay=delay)
            message.update({'submitted':True,'exec':execut})
            message.update({"median": median(scenario_values),
                            "mean": np.mean(scenario_values),
                            "std": np.std(scenario_values),
                            "values[0]": scenario_values[0],
                            "values[-1]": scenario_values[-1]})

            num_intervals = self.update_frequency(name=name, delay=delay)
            predict_time, dt, earliest, latest, expected_at = self.estimate_prediction_time(lagged_times=lagged_times, delay=delay, num_intervals=num_intervals)
            message.update({'earliest': earliest, 'latest': latest})
            message.update({'expected_at':expected_at})

            if predict_time>time.time():
                message.update({'t_next':predict_time})
                self.next_prediction_time[horizon] = predict_time
            else:
                if horizon in self.backoff:
                    self.next_prediction_time[horizon] = time.time() + self.backoff[horizon]
                    self.backoff[horizon] = 2*self.backoff[horizon]
                else:
                    self.next_prediction_time[horizon] = time.time()+ delay
                    self.backoff[horizon] = delay
                message.update({'backoff':self.backoff[horizon]})

            if not execut:
                message.update({"submitted":False,"reason":"execution failure","confirms":self.get_confirms()[-1:], "errors": self.get_errors()[-1:]})
                print("---------- ERROR ------------")
                pprint.pprint(message)
                print("-----------------------------")

        if self.feel_like_talking():
            message.update({'balance': self.get_balance(),
                            "errors":self.get_errors()[-10:]})
            print(" ")
            pprint.pprint(message)
            print(" ", flush=True)
        return execut

    def initial_urgency_multiplier(self,horizon):
        """ Seconds """
        return 20 if '~' in horizon else 1

    def run(self,timeout=None):
        # Pick up where we left off, but stagger
        self.performance = self.get_performance()
        self.active = self.get_active()
        self.start_time = time.time()
        self.end_time = time.time()+timeout if timeout is not None else time.time()+10000000
        print('Currently predicting for ' + str(len(self.active)) + ' horizons')
        self.stream_candidates = self.candidate_streams()
        print('Found ' + str(len(self.stream_candidates)) + ' candidate streams.', flush=True)
        desired_streams = [self.horizon_name(stream_name, horizon) for stream_name in self.stream_candidates for horizon in [70, 310, 910]]
        self.next_prediction_time = dict( [ (stream, time.time() + k*self.initial_urgency_multiplier(stream)) \
            for k, stream in enumerate(self.active) if stream in desired_streams])


        print("---------- Restarting -------------")
        pprint.pprint(self.__repr__())
        print("-----------------------------------",flush=True)


        while time.time()<self.end_time:

            # Periodically consider entering new horizons and withdrawing from others
            self.update_seconds_until_next()
            if self.seconds_until_next>np.random.randn()*120:
                if time.time()-self.last_performance_check>60*60:
                    self.performance = self.get_performance()
                    self.stream_candidates = self.candidate_streams()
                    print('Found '+str(len(self.stream_candidates))+ ' candidate streams.')
                    self.active = self.get_active()
                    print('Currently predicting for ' + str(len(self.active)) + ' horizons')
                    self.withdraw_from_worst()
                    self.last_performance_check = time.time()

            # If there is time consider entering a new stream, but not too often
            self.update_seconds_until_next()
            if (self.seconds_until_next>np.random.rand()*500) or (len(self.active)<10):
                if (time.time()-self.last_new_horizon>10*60) or (len(self.active)==0):
                    horizon = self.next_horizon(exclude=self.horizon_blacklist)
                    if horizon:
                        name, delay = self.split_horizon_name(horizon)
                        lagged_times = self.get_lagged_times(name=name)
                        self.predict_and_submit(name=name, delay=delay, lagged_times=lagged_times)
                        print(" ", )
                        print(" Found something new to predict: " + horizon)
                        print(" ",flush=True )
                        self.last_new_horizon = time.time()
                        if self.feel_like_talking():
                            print(" ")
                            pprint.pprint(self.recent_updates())
                            print(" ", flush=True)
                    else:
                        print(" ")
                        print(" Couldn't find anything else to predict, for now. ",flush=True )
                        print(" ",flush=True)

            # If there is time, maybe we chill, or do something productive like offline estimation
            self.update_seconds_until_next()
            if self.seconds_until_next>5:
                print('Downtime for '+str(self.seconds_until_next)+' seconds. Next '+ str(self.upcoming(num=1)))
                self.downtime(seconds=self.seconds_until_next-1)
                self.update_seconds_until_next()

            if self.seconds_until_next>0:
                time.sleep(self.seconds_until_next)

            # It might be time to submit predictions
            for horizon, seconds_to_go in self.upcoming(num=1,relative=True):
                name, delay = self.split_horizon_name(horizon)
                if seconds_to_go<1:
                    if seconds_to_go>0:
                        time.sleep(seconds_to_go+0.1)
                    lagged_times = self.get_lagged_times(name=name)
                    if lagged_times and (abs(time.time()-lagged_times[0])<30):
                        # Received recent data, so it is time to predict
                        name, delay = self.split_horizon_name(horizon)
                        self.predict_and_submit(name=name, delay=delay, lagged_times=lagged_times )
                    else:
                        time.sleep(1)




